/*This code shows how MDMA descriptor based mode can be used to emulate circular buffer memory
 * to memory DMA transfer mode supported by EMDMA to achieve higher performance.
  */

#include "main.h"
#include <sys/platform.h>
#include <sys/adi_core.h>
#include "adi_initialize.h"
#include "math.h"

#pragma align 32
uint32_t uiSrcBuff1[BUFF_SIZE];

#pragma align 32
uint32_t uiDestBuff1[BUFF_SIZE];

uint32_t uiSrcTCB1[2*TCB_SIZE];
uint32_t uiDstTCB1[2*TCB_SIZE];


volatile clock_t clock_stop;
volatile clock_t clock_start;
volatile int EMDMA_Normal_Write_Cycles;
volatile int EMDMA_Normal_Read_Cycles;
volatile int EMDMA_Circular_Write_Cycles;
volatile int EMDMA_Circular_Read_Cycles;

volatile int MDMA_MSIZE4_Write_Cycles;
volatile int MDMA_MSIZE4_Read_Cycles;
volatile int MDMA_MSIZE32_Write_Cycles;
volatile int MDMA_MSIZE32_Read_Cycles;

volatile int MDMA_MSIZE4_Circ_Write_Cycles;
volatile int MDMA_MSIZE4_Circ_Read_Cycles;
volatile int MDMA_MSIZE32_Circ_Write_Cycles;
volatile int MDMA_MSIZE32_Circ_Read_Cycles;


void Buff_Init(unsigned int* pBuff, unsigned int count);
void Mem_Copy_MDMA_MSIZE4_Circ_Write(unsigned int srcadd, unsigned int dstadd, unsigned int count, unsigned int dstaddbase, unsigned int buflen);

int main(void)
{
	int failure=0; int i;

	//Initialize the source buffer with incremental data
	Buff_Init(uiSrcBuff1, BUFF_SIZE);

	//Configure MDMA3 and EMDMA as secure masters
	*pREG_SPU0_SECUREP141=0x3;
	*pREG_SPU0_SECUREP158=0x3;

	//Perform write and read circular buffer DMA to/from DDR3 memory using EMDMA
	#ifdef  ENABLE_EMDMA_CIRCULAR

		//Write circular EMDMA
		clock_start=clock();
		EMDMA_CIRCULAR_CONFIG(0,EMDMA_CTL_VALUE_WRITE,((uint32_t)uiSrcBuff1>>2)|0xA000000, 1, BUFF_SIZE, (DMC0_START_ADDRESS+INDEX_OFFSET)>>2, 1, BUFF_SIZE,DMC0_START_ADDRESS>>2,BUFF_SIZE);
		WAIT_FOR_EMDMA_OVER(0);
		clock_stop=clock();
		EMDMA_Circular_Write_Cycles= clock_stop - clock_start;
		printf("\n\nEMDMA circular write cycles = %d",EMDMA_Circular_Write_Cycles );

		DISABLE_EMDMA(0);

		//Read circular EMDMA
		clock_start=clock();
		EMDMA_CIRCULAR_CONFIG(0,EMDMA_CTL_VALUE_READ,((uint32_t)uiDestBuff1>>2)|0xA000000, 1, BUFF_SIZE, (DMC0_START_ADDRESS+INDEX_OFFSET)>>2, 1, BUFF_SIZE,DMC0_START_ADDRESS>>2,BUFF_SIZE);
		WAIT_FOR_EMDMA_OVER(0);
		clock_stop=clock();
		EMDMA_Circular_Read_Cycles= clock_stop - clock_start;
		printf("\n\nEMDMA circular read cycles = %d",EMDMA_Circular_Read_Cycles );

		DISABLE_EMDMA(0);

		//Compare the source and destination buffers to ensure that the write and read DMA worked as expected
		for(i=0;i<BUFF_SIZE;i++)
		{
			if(uiSrcBuff1[i]!=uiDestBuff1[i])
				failure++;
		}

		if(failure==0)
			printf("\n\nEMDMA write and read passed...\n\n");
		else
			printf("\n\nEMDMA write and read failed...\n\n");
	#endif

	//Perform write and read circular buffer DMA to/from DDR3 memory using MDMA with MSIZE=4 bytes
	#ifdef ENABLE_MDMA_MSIZE4_CIRCULAR

		//Write circular MDMA
		clock_start=clock();
		Mem_Copy_MDMA_MSIZE4_Circ_Write(((uint32_t)uiSrcBuff1|0x28000000), DMC0_START_ADDRESS+INDEX_OFFSET*4, BUFF_SIZE, DMC0_START_ADDRESS, BUFF_SIZE);
		clock_stop=clock();
		MDMA_MSIZE4_Circ_Write_Cycles= clock_stop - clock_start;
		printf("\n\nMDMA MSIZE4 circular write cycles = %d", MDMA_MSIZE4_Circ_Write_Cycles );
		DISABLE_MDMA(43,44);

		//Read circular EMDMA
		clock_start=clock();
		Mem_Copy_MDMA_MSIZE4_Circ_Read(DMC0_START_ADDRESS+INDEX_OFFSET*4, ((uint32_t)uiSrcBuff1|0x28000000), BUFF_SIZE, DMC0_START_ADDRESS, BUFF_SIZE);
		clock_stop=clock();
		MDMA_MSIZE4_Circ_Read_Cycles= clock_stop - clock_start;
		printf("\n\nMDMA MSIZE4 circular Read cycles = %d", MDMA_MSIZE4_Circ_Read_Cycles );
		DISABLE_MDMA(43,44);

		//Compare the source and destination buffers to ensure that the write and read DMA worked as expected
		for(i=0;i<BUFF_SIZE;i++)
		{
			if(uiSrcBuff1[i]!=uiDestBuff1[i])
				failure++;
		}

		if(failure==0)
			printf("\n\nMDMA MSIZE4 write and read passed...\n\n");
		else
			printf("\n\nMDMA MSIZE4 write and read failed...\n\n");

	#endif

	//Perform write and read circular buffer DMA to/from DDR3 memory using MDMA with MSIZE=4 bytes
	#ifdef ENABLE_MDMA_MSIZE32_CIRCULAR

		//Write circular MDMA
		clock_start=clock();
		Mem_Copy_MDMA_MSIZE32_Circ_Write(((uint32_t)uiSrcBuff1|0x28000000), DMC0_START_ADDRESS+INDEX_OFFSET*4, BUFF_SIZE, DMC0_START_ADDRESS, BUFF_SIZE);
		clock_stop=clock();
		MDMA_MSIZE32_Circ_Write_Cycles= clock_stop - clock_start;
		printf("\n\nMDMA MSIZE32 circular write cycles = %d", MDMA_MSIZE32_Circ_Write_Cycles );
		DISABLE_MDMA(43,44);

		//Read circular EMDMA
		clock_start=clock();
		Mem_Copy_MDMA_MSIZE32_Circ_Read(DMC0_START_ADDRESS+INDEX_OFFSET*4, ((uint32_t)uiSrcBuff1|0x28000000), BUFF_SIZE, DMC0_START_ADDRESS, BUFF_SIZE);
		clock_stop=clock();
		MDMA_MSIZE32_Circ_Read_Cycles= clock_stop - clock_start;
		printf("\n\nMDMA MSIZE32 circular Read cycles = %d", MDMA_MSIZE32_Circ_Read_Cycles );
		DISABLE_MDMA(43,44);

		//Compare the source and destination buffers to ensure that the write and read DMA worked as expected
		for(i=0;i<BUFF_SIZE;i++)
		{
			if(uiSrcBuff1[i]!=uiDestBuff1[i])
				failure++;
		}

		if(failure==0)
			printf("\n\nMDMA MSIZE32 write and read passed...\n\n");
		else
			printf("\n\nMDMA MSIZE32 write and read failed...\n\n");

	#endif

	return 0;
}

void Buff_Init(unsigned int* pBuff, unsigned int count)
{
	int i;

	for(i=0;i<count;i++)
		pBuff[i]=i;
}

// API to emulate circular buffer MDMA write with MSIZE=4 bytes
void Mem_Copy_MDMA_MSIZE4_Circ_Write(unsigned int srcadd, unsigned int dstadd, unsigned int count, unsigned int dstaddbase, unsigned int buflen)
{
	unsigned int count1, count2;

	//Split the MDMA into two descriptor based MDMA transfers
	count1=((dstaddbase+buflen*4)-dstadd)/4;
	count2=count-count1;

	uiSrcTCB1[0]=srcadd;
	uiSrcTCB1[1]=SRC_DMA_CFG5;
	uiSrcTCB1[2]=count;
	uiSrcTCB1[3]=4;

	uiDstTCB1[0]=dstadd;
	uiDstTCB1[1]=DST_DMA_CFG3;
	uiDstTCB1[2]=count1;
	uiDstTCB1[3]=4;
	uiDstTCB1[4]=dstaddbase;
	uiDstTCB1[5]=DST_DMA_CFG5;
	uiDstTCB1[6]=count2;
	uiDstTCB1[7]=4;

	*pREG_DMA43_DSCPTR_CUR=(uint32_t)uiSrcTCB1|0x28000000;
	*pREG_DMA44_DSCPTR_CUR=(uint32_t)uiDstTCB1|0x28000000;
	*pREG_DMA44_CFG=SRC_DMA_CFG3;
	*pREG_DMA43_CFG=SRC_DMA_CFG3;

	WAIT_FOR_DMADONE(44);
}

// API to emulate circular buffer MDMA read with MSIZE=4 bytes
void Mem_Copy_MDMA_MSIZE4_Circ_Read(unsigned int srcadd, unsigned int dstadd, unsigned int count, unsigned int srcaddbase, unsigned int buflen)
{
	unsigned int count1, count2;

	//Split the MDMA into two descriptor based MDMA transfers
	count1=((srcaddbase+buflen*4)-srcadd)/4;
	count2=count-count1;

	uiDstTCB1[0]=dstadd;
	uiDstTCB1[1]=DST_DMA_CFG5;
	uiDstTCB1[2]=count;
	uiDstTCB1[3]=4;

	uiSrcTCB1[0]=srcadd;
	uiSrcTCB1[1]=SRC_DMA_CFG3;
	uiSrcTCB1[2]=count1;
	uiSrcTCB1[3]=4;
	uiSrcTCB1[4]=srcaddbase;
	uiSrcTCB1[5]=SRC_DMA_CFG5;
	uiSrcTCB1[6]=count2;
	uiSrcTCB1[7]=4;

	*pREG_DMA43_DSCPTR_CUR=(uint32_t)uiSrcTCB1|0x28000000;
	*pREG_DMA44_DSCPTR_CUR=(uint32_t)uiDstTCB1|0x28000000;
	*pREG_DMA44_CFG=SRC_DMA_CFG3;
	*pREG_DMA43_CFG=SRC_DMA_CFG3;

	WAIT_FOR_DMADONE(44);
}

// API to emulate circular buffer MDMA write with MSIZE=32 bytes
void Mem_Copy_MDMA_MSIZE32_Circ_Write(unsigned int srcadd, unsigned int dstadd, unsigned int count, unsigned int dstaddbase, unsigned int buflen)
{
	unsigned int count1, count2;

	//Split the MDMA into two descriptor based MDMA transfers
	count1=((dstaddbase+buflen*4)-dstadd)/4;
	count2=count-count1;

	uiSrcTCB1[0]=srcadd;
	uiSrcTCB1[1]=SRC_DMA_CFG6;
	uiSrcTCB1[2]=count/8;
	uiSrcTCB1[3]=32;

	uiDstTCB1[0]=dstadd;
	uiDstTCB1[1]=DST_DMA_CFG4;
	uiDstTCB1[2]=count1/8;
	uiDstTCB1[3]=32;
	uiDstTCB1[4]=dstaddbase;
	uiDstTCB1[5]=DST_DMA_CFG6;
	uiDstTCB1[6]=count2/8;
	uiDstTCB1[7]=32;

	*pREG_DMA43_DSCPTR_CUR=(uint32_t)uiSrcTCB1|0x28000000;
	*pREG_DMA44_DSCPTR_CUR=(uint32_t)uiDstTCB1|0x28000000;
	*pREG_DMA44_CFG=SRC_DMA_CFG3;
	*pREG_DMA43_CFG=SRC_DMA_CFG3;

	WAIT_FOR_DMADONE(44);
}

// API to emulate circular buffer MDMA read with MSIZE=32 bytes
void Mem_Copy_MDMA_MSIZE32_Circ_Read(unsigned int srcadd, unsigned int dstadd, unsigned int count, unsigned int srcaddbase, unsigned int buflen)
{
	unsigned int count1, count2;

	//Split the MDMA into two descriptor based MDMA transfers
	count1=((srcaddbase+buflen*4)-srcadd)/4;
	count2=count-count1;

	uiDstTCB1[0]=dstadd;
	uiDstTCB1[1]=DST_DMA_CFG6;
	uiDstTCB1[2]=count/8;
	uiDstTCB1[3]=32;

	uiSrcTCB1[0]=srcadd;
	uiSrcTCB1[1]=SRC_DMA_CFG4;
	uiSrcTCB1[2]=count1/8;
	uiSrcTCB1[3]=32;
	uiSrcTCB1[4]=srcaddbase;
	uiSrcTCB1[5]=SRC_DMA_CFG6;
	uiSrcTCB1[6]=count2/8;
	uiSrcTCB1[7]=32;

	*pREG_DMA43_DSCPTR_CUR=(uint32_t)uiSrcTCB1|0x28000000;
	*pREG_DMA44_DSCPTR_CUR=(uint32_t)uiDstTCB1|0x28000000;
	*pREG_DMA44_CFG=SRC_DMA_CFG3;
	*pREG_DMA43_CFG=SRC_DMA_CFG3;

	WAIT_FOR_DMADONE(44);
}
